/*
 * Sone - SoneTextParserTest.java - Copyright © 2011–2013 David Roden
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package net.pterodactylus.sone.text;

import static org.hamcrest.MatcherAssert.assertThat;
import static org.hamcrest.Matchers.is;
import static org.hamcrest.Matchers.notNullValue;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;

import java.io.IOException;
import java.io.StringReader;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import net.pterodactylus.sone.data.Post;
import net.pterodactylus.sone.data.Sone;
import net.pterodactylus.sone.data.impl.IdOnlySone;
import net.pterodactylus.sone.database.PostProvider;
import net.pterodactylus.sone.database.SoneProvider;

import com.google.common.base.Function;
import com.google.common.base.Optional;
import org.junit.Test;
import org.mockito.Mockito;
import org.mockito.stubbing.OngoingStubbing;

/**
 * JUnit test case for {@link SoneTextParser}.
 *
 * @author <a href="mailto:bombe@pterodactylus.net">David ‘Bombe’ Roden</a>
 */
public class SoneTextParserTest {

	private final SoneProvider soneProvider = new TestSoneProvider();
	private final TestPostProvider postProvider = new TestPostProvider();
	private final SoneTextParser soneTextParser = new SoneTextParser(soneProvider, postProvider);

	/**
	 * Tests basic plain-text operation of the parser.
	 *
	 * @throws IOException
	 *             if an I/O error occurs
	 */
	@Test
	public void testPlainText() throws IOException {
		Iterable<Part> parts;

		/* check basic operation. */
		parts = soneTextParser.parse(null, new StringReader("Test."));
		assertThat(parts, notNullValue());
		assertThat(convertText(parts, PlainTextPart.class), is("Test."));

		/* check empty lines at start and end. */
		parts = soneTextParser.parse(null, new StringReader("\nTest.\n\n"));
		assertThat(parts, notNullValue());
		assertThat(convertText(parts, PlainTextPart.class), is("Test."));

		/* check duplicate empty lines in the text. */
		parts = soneTextParser.parse(null, new StringReader("\nTest.\n\n\nTest."));
		assertThat(parts, notNullValue());
		assertThat(convertText(parts, PlainTextPart.class), is("Test.\n\nTest."));
	}

	/**
	 * Tests parsing of KSK links.
	 *
	 * @throws IOException
	 *             if an I/O error occurs
	 */
	@Test
	public void testKSKLinks() throws IOException {
		Iterable<Part> parts;

		/* check basic links. */
		parts = soneTextParser.parse(null, new StringReader("KSK@gpl.txt"));
		assertThat(parts, notNullValue());
		assertThat(convertText(parts, FreenetLinkPart.class), is("[KSK@gpl.txt|gpl.txt|gpl.txt]"));

		/* check embedded links. */
		parts = soneTextParser.parse(null, new StringReader("Link is KSK@gpl.txt\u200b."));
		assertThat(parts, notNullValue());
		assertThat(convertText(parts, PlainTextPart.class, FreenetLinkPart.class), is("Link is [KSK@gpl.txt|gpl.txt|gpl.txt]\u200b."));

		/* check embedded links and line breaks. */
		parts = soneTextParser.parse(null, new StringReader("Link is KSK@gpl.txt\nKSK@test.dat\n"));
		assertThat(parts, notNullValue());
		assertThat(convertText(parts, PlainTextPart.class, FreenetLinkPart.class),
			is("Link is [KSK@gpl.txt|gpl.txt|gpl.txt]\n[KSK@test.dat|test.dat|test.dat]"));
	}

	/**
	 * Test case for a bug that was discovered in 0.6.7.
	 *
	 * @throws IOException
	 *             if an I/O error occurs
	 */
	@Test
	public void testEmptyLinesAndSoneLinks() throws IOException {
		Iterable<Part> parts;

		/* check basic links. */
		parts = soneTextParser.parse(null, new StringReader("Some text.\n\nLink to sone://DAxKQzS48mtaQc7sUVHIgx3fnWZPQBz0EueBreUVWrU and stuff."));
		assertThat(parts, notNullValue());
		assertThat(convertText(parts, PlainTextPart.class, SonePart.class),
			is("Some text.\n\nLink to [Sone|DAxKQzS48mtaQc7sUVHIgx3fnWZPQBz0EueBreUVWrU] and stuff."));
	}

	/**
	 * Test for a bug discovered in Sone 0.8.4 where a plain “http://” would be
	 * parsed into a link.
	 *
	 * @throws IOException
	 *             if an I/O error occurs
	 */
	@Test
	public void testEmpyHttpLinks() throws IOException {
		Iterable<Part> parts;

		/* check empty http links. */
		parts = soneTextParser.parse(null, new StringReader("Some text. Empty link: http:// – nice!"));
		assertThat(parts, notNullValue());
		assertThat(convertText(parts, PlainTextPart.class), is("Some text. Empty link: http:// – nice!"));
	}

	@Test
	public void linksToPostAreParsedCorrectly() throws IOException {
		postProvider.addValidPostId("foo", "Post about foo...");
		Iterable<Part> parts = soneTextParser.parse(null, new StringReader("This post://foo is awesome."));
		assertThat(convertText(parts, PlainTextPart.class, PostPart.class), is("This [post|foo|Post about foo...] is awesome."));
	}

	//
	// PRIVATE METHODS
	//

	/**
	 * Converts all given {@link Part}s into a string, validating that the
	 * part’s classes match only the expected classes.
	 *
	 * @param parts
	 *            The parts to convert to text
	 * @param validClasses
	 *            The valid classes; if no classes are given, all classes are
	 *            valid
	 * @return The converted text
	 */
	private static String convertText(Iterable<Part> parts, Class<?>... validClasses) {
		StringBuilder text = new StringBuilder();
		for (Part part : parts) {
			assertThat(part, notNullValue());
			boolean classValid = validClasses.length == 0;
			for (Class<?> validClass : validClasses) {
				if (validClass.isAssignableFrom(part.getClass())) {
					classValid = true;
					break;
				}
			}
			assertThat("Part’s Class (" + part.getClass() + ") is not one of " + Arrays.toString(validClasses), classValid, is(true));
			if (part instanceof PlainTextPart) {
				text.append(part.getText());
			} else if (part instanceof FreenetLinkPart) {
				FreenetLinkPart freenetLinkPart = (FreenetLinkPart) part;
				text.append('[')
					.append(freenetLinkPart.getLink())
					.append('|')
					.append(freenetLinkPart.isTrusted() ? "trusted|" : "")
					.append(freenetLinkPart.getTitle())
					.append('|')
					.append(freenetLinkPart.getText())
					.append(']');
			} else if (part instanceof LinkPart) {
				LinkPart linkPart = (LinkPart) part;
				text.append('[')
					.append(linkPart.getLink())
					.append('|')
					.append(linkPart.getTitle())
					.append('|')
					.append(linkPart.getText())
					.append(']');
			} else if (part instanceof SonePart) {
				SonePart sonePart = (SonePart) part;
				text.append("[Sone|").append(sonePart.getSone().getId()).append(']');
			} else if (part instanceof PostPart) {
				PostPart postPart = (PostPart) part;
				text.append("[post|").append(postPart.getPost().getId()).append('|').append(postPart.getPost().getText()).append(']');
			}
		}
		return text.toString();
	}

	/**
	 * Mock Sone provider.
	 *
	 * @author <a href="mailto:bombe@pterodactylus.net">David ‘Bombe’ Roden</a>
	 */
	private static class TestSoneProvider implements SoneProvider {

		@Override
		public Function<String, Optional<Sone>> soneLoader() {
			return new Function<String, Optional<Sone>>() {
				@Override
				public Optional<Sone> apply(String soneId) {
					return getSone(soneId);
				}
			};
		}

		/**
		 * {@inheritDoc}
		 */
		@Override
		public Optional<Sone> getSone(final String soneId) {
			return Optional.<Sone>of(new IdOnlySone(soneId));
		}

		/**
		 * {@inheritDocs}
		 */
		@Override
		public Collection<Sone> getSones() {
			return null;
		}

		/**
		 * {@inheritDocs}
		 */
		@Override
		public Collection<Sone> getLocalSones() {
			return null;
		}

		/**
		 * {@inheritDocs}
		 */
		@Override
		public Collection<Sone> getRemoteSones() {
			return null;
		}

	}

	private static class TestPostProvider implements PostProvider {

		private final Map<String, String> posts = new HashMap<String, String>();

		private void addValidPostId(String validPostId, String text) {
			posts.put(validPostId, text);
		}

		@Override
		public Collection<Post> getDirectedPosts(String recipientId) {
			return Collections.emptyList();
		}

		@Override
		public Collection<Post> getPosts(String soneId) {
			return Collections.emptyList();
		}

		@Override
		public Optional<Post> getPost(String postId) {
			if (posts.containsKey(postId)) {
				Post post = mock(Post.class);
				when(post.getId()).thenReturn(postId);
				when(post.getText()).thenReturn(posts.get(postId));
				return Optional.of(post);
			}
			return Optional.absent();
		}

	}

}
